#include <Arduino.h>
#include <Wire.h>
#include <Adafruit_GFX.h>
#include <Adafruit_SSD1306.h>
#undef WHITE
#undef BLACK
#include <SparkFunTMP102.h>
#include <Adafruit_NeoPixel.h>

#include "touch.h"

enum Pinout {
  DRIVE2 = 0,
  DRIVE1 = 1,
  LED_W = 2,
  WS2812_DISPLAY = 2,
  LED_Y = 3,
  EN_MEASURE_CURRENT = 3,
  LED_B = 4,
  LED_G = 5,
  LED_R = 6,
  MATRIX_IN1 = 7,
  SBU1 = 8,
  SBU2 = 9,
  VBUS_DET = 10,
  BOOT2 = 11,
  I2C_OLED_SDA = 12,
  I2C_OLED_SCL = 13,
  WS2811 = 14,
  TX_EN = 15,
  TX = 16,
  RX = 17,
  MATRIX_IN2 = 18,
  REED3 = 19,
  REED4 = 20,
  REED2 = 21,
  REED1 = 22,
  MATRIX_IN3 = 23,
  DIGOUT1 = 24,
  DIGOUT2 = 25,
  ANALOG_IN1 = 26,
  CURRENT2 = 27,
  CURRENT1 = 28,
  MEASURE_VCC = 29,
};

const int SCREEN_ADDRESS = 0x3C;
Adafruit_SSD1306 display(128, 64, &Wire, -1);
bool hasDisplay;

#include "logo.h"

TMP102 sensor0, sensor1;

HeizungTouch touch;
Adafruit_NeoPixel pixelsDisplay(60, WS2812_DISPLAY, NEO_GRB + NEO_KHZ800);

static bool initial = true;

void setup() {
  rp2040.enableDoubleResetBootloader();

  pinMode(DIGOUT1, OUTPUT);
  digitalWrite(DIGOUT1, LOW);

  Serial.begin(9600);  // USB serial, baudrate is ignored
  if (0) {
    while (!Serial)
      ;
    Serial.println("a1");
  }

  pinMode(DRIVE1, OUTPUT); digitalWrite(DRIVE1, LOW);
  pinMode(DRIVE2, OUTPUT); digitalWrite(DRIVE2, LOW);
  pinMode(WS2812_DISPLAY, OUTPUT);
  pinMode(3, OUTPUT);  digitalWrite(EN_MEASURE_CURRENT, LOW);
  pinMode(4, OUTPUT);
  pinMode(5, OUTPUT);
  pinMode(6, OUTPUT);

  pinMode(BOOT2, INPUT_PULLUP);
  pinMode(WS2811, INPUT_PULLUP);
  pinMode(REED1, OUTPUT);
  pinMode(REED2, OUTPUT);
  pinMode(REED3, OUTPUT);
  pinMode(REED4, OUTPUT);
  pinMode(MATRIX_IN1, INPUT_PULLUP);
  pinMode(MATRIX_IN3, INPUT_PULLUP);

  pinMode(TX_EN, OUTPUT);
  digitalWrite(TX_EN, LOW);

  pixelsDisplay.begin();
  pixelsDisplay.clear();
  pixelsDisplay.show();

  Serial1.setTX(TX);  // UART0
  Serial1.setRX(RX);
  Wire.setSDA(I2C_OLED_SDA);  // I2C0
  Wire.setSCL(I2C_OLED_SCL);

  Serial1.begin(115200);
  Wire.begin();
  Wire.setClock(100000);

  display.setRotation(2);
  hasDisplay = display.begin(SSD1306_SWITCHCAPVCC, SCREEN_ADDRESS);
  if (!hasDisplay) {
    Serial.println(F("SSD1306 allocation failed"));
  } else {
    display.clearDisplay();
    display.drawBitmap(0, 0, logo_data, logo_width, logo_height, SSD1306_WHITE);
    display.display();
  }

  if (sensor0.begin(0x48)) {
    //0-3: 0:0.25Hz, 1:1Hz, 2:4Hz, 3:8Hz
    sensor0.setConversionRate(2);
    sensor0.setExtendedMode(1);
    sensor0.wakeup();
  }
  if (sensor1.begin(0x49)) {
    //0-3: 0:0.25Hz, 1:1Hz, 2:4Hz, 3:8Hz
    sensor1.setConversionRate(2);
    sensor1.setExtendedMode(1);
    sensor1.wakeup();
  }

  touch.init();
}

void displayPrintHexFixed(uint32_t x, int width) {
  while (width > 1 && (x >> (4*(width-1))) == 0) {
    display.write('0');
    width--;
  }
  display.print(x, HEX);
}

void readButtons(bool buttons[4]) {
  // Button Matrix:
  // 1: SDA, IN1
  // 2: SDA, IN3
  // 3: SCL, IN1
  // 4: SCL, IN4
  Wire.end();
  pinMode(I2C_OLED_SDA, OUTPUT);
  pinMode(I2C_OLED_SCL, INPUT_PULLUP);
  delayMicroseconds(1);
  buttons[0] = !digitalRead(MATRIX_IN1);
  buttons[1] = !digitalRead(MATRIX_IN3);
  pinMode(I2C_OLED_SDA, INPUT_PULLUP);
  pinMode(I2C_OLED_SCL, OUTPUT);
  delayMicroseconds(1);
  buttons[2] = !digitalRead(MATRIX_IN1);
  buttons[3] = !digitalRead(MATRIX_IN3);
  pinMode(I2C_OLED_SDA, INPUT_PULLUP);
  pinMode(I2C_OLED_SCL, INPUT_PULLUP);
  Wire.begin();
}

// show remaining time in milliseconds and use short timeout for 3rd button
constexpr bool DEBUG_TIME = false;
// show touch bits on display
constexpr bool DEBUG_TOUCH = true;
// make LEDs a lot darker so it is easier to take a photo
constexpr bool LEDS_FOR_PHOTO = false;

typedef uint16_t temperature_t;
constexpr temperature_t TEMPERATURE_MIN = 100;  // 10.0 deg C
constexpr temperature_t TEMPERATURE_MAX = 300;
constexpr temperature_t TEMPERATURE_STANDBY = 150;
typedef unsigned long millis_t;
typedef signed long millis_diff_t;
constexpr millis_t MINUTES = 60*1000;
constexpr millis_t HOURS = 3600*1000;
constexpr millis_t TIME_UNTIL_STANDBY_DEFAULT = 2*HOURS + MINUTES/2;
constexpr millis_t TIME_UNTIL_STANDBY_DEFAULT2 = DEBUG_TIME ? 10*1000 : TIME_UNTIL_STANDBY_DEFAULT;
constexpr millis_t TIME_UNTIL_STANDBY_MAX = 48*HOURS - MINUTES/2;
constexpr millis_t TIME_DISABLED = 0;

void loop() {
  static uint16_t wantedTemperature = TEMPERATURE_STANDBY;
  static long wantedTemperatureUntil = TIME_DISABLED;
  bool changed = false;

  // Calculate diff with unsigned numbers because signed overflow is undefined behaviour
  // but convert the result to signed so we can compare it to zero (rather than LONG_MAX/2).
  const millis_t now = millis();
  const millis_diff_t time_to_standby = (millis_diff_t)(wantedTemperatureUntil - now);
  if (time_to_standby <= 0) {
    wantedTemperatureUntil = 0;
    wantedTemperature = TEMPERATURE_STANDBY;
  }

  static int button_timer = 0;
  bool buttons[4];
  if (button_timer < 500)
    button_timer++;
  else {
    readButtons(buttons);

    if (buttons[0] && wantedTemperature > TEMPERATURE_MIN) {
      wantedTemperature--;
      changed = true;
    }
    if (buttons[1] && wantedTemperature < TEMPERATURE_MAX) {
      wantedTemperature++;
      changed = true;
      if (wantedTemperature > TEMPERATURE_STANDBY && (wantedTemperatureUntil == 0 || time_to_standby < TIME_UNTIL_STANDBY_DEFAULT)) {
        wantedTemperatureUntil = now + TIME_UNTIL_STANDBY_DEFAULT;
        if (wantedTemperatureUntil == TIME_DISABLED)
          wantedTemperatureUntil++;  // avoid special value
      }
    }
    if (buttons[2] && wantedTemperature > TEMPERATURE_STANDBY) {
      if (0)
        wantedTemperatureUntil = now + TIME_UNTIL_STANDBY_DEFAULT;
      else
        // very small timeout for debugging
        wantedTemperatureUntil = now + TIME_UNTIL_STANDBY_DEFAULT2;

      if (wantedTemperatureUntil == TIME_DISABLED)
        wantedTemperatureUntil++;  // avoid special value
    }
    if (buttons[3] && wantedTemperatureUntil != TIME_DISABLED) {
      changed = true;
      if (time_to_standby >= TIME_UNTIL_STANDBY_MAX)
        ;
      else if (time_to_standby < 1*HOURS)
        wantedTemperatureUntil += 1 * MINUTES;
      else if (time_to_standby < 2*HOURS)
        wantedTemperatureUntil += 2 * MINUTES;
      else if (time_to_standby < 10*HOURS)
        wantedTemperatureUntil += 5 * MINUTES;
      else
        wantedTemperatureUntil += 20 * MINUTES;

      if (wantedTemperatureUntil == TIME_DISABLED)
        wantedTemperatureUntil++;  // avoid special value
    }
  }

  //FIXME use touch areas for input (double-tap or slide to set temperature)
  touch.update(now);
  if (DEBUG_TOUCH)
    changed |= touch.hasChanged();

  auto temp0 = sensor0.readTempC();
  auto temp1 = sensor1.readTempC();
  auto temp = !isnan(temp1) ? temp1 : temp0;
  auto actualTemperature = (temperature_t)(temp * 10);

  // very stupid control loop but should be good enough, for now
  // (Good control loop is not part of minimum viable product and can later be done in Python if anyone cares to do it.)
  bool heatOn = (actualTemperature < wantedTemperature);
  digitalWrite(DRIVE1, heatOn);
  digitalWrite(DRIVE2, heatOn);

  static int display_timer = 0;
  if (display_timer < 500 && !changed)
    display_timer++;
  else {
    display_timer = 0;
  }

  if (changed || initial || display_timer == 0) {
    initial = false;

    auto BLACK  = pixelsDisplay.Color(0, 0, 0);
    auto WHITE  = pixelsDisplay.Color(8, 8, 8);
    auto BLUE   = pixelsDisplay.Color(0, 0, 24);
    auto GREEN  = pixelsDisplay.Color(0, 8, 0);
    auto ORANGE = pixelsDisplay.Color(24, 16, 0);

    if (LEDS_FOR_PHOTO && 0) {
      WHITE  = pixelsDisplay.Color(3, 3, 2);
      ORANGE = pixelsDisplay.Color(12, 6, 0);
    } else if (LEDS_FOR_PHOTO) {
      WHITE  = pixelsDisplay.Color(8, 8, 4);
    }

    int pixelsOnWanted = 256 + (wantedTemperature - TEMPERATURE_MIN) * 256 * (60-16) / (TEMPERATURE_MAX - TEMPERATURE_MIN);
    int pixelsOnActual = 256 + (actualTemperature - TEMPERATURE_MIN) * 256 * (60-16) / (TEMPERATURE_MAX - TEMPERATURE_MIN);
    for (int i=0; i<pixelsDisplay.numPixels(); i++) {
      // actual ring starts on the right and goes counter-clockwise ("mathematical angle")
      // but we want from lower left, clockwise to lower right
      int j;
      if (i < 38)
        j = 37-i;
      else if (i < 38+15)
        j = 100;  // lower part
      else
        j = 38 + (59-i);

      bool wanted = (j < pixelsOnWanted/256);
      int actual = (pixelsOnActual - 256*j);
      auto color = BLACK;
      if (j == 100)
        color = BLACK;  // lower part of the ring
      else if (wanted && actual > 0)
        color = WHITE;  // actual and wanted
      else if (wanted && actual <= 0) {
        color = ORANGE; // too cold -> needs heating
        //TODO blend with white
      } else if (!wanted && actual > 0 && wantedTemperatureUntil == TIME_DISABLED)
        color = GREEN; // too hot but for standby temperature
      else if (!wanted && actual > 0) {
        color = BLUE; // too hot -> needs cooling
        if (actual < 256)
          color = pixelsDisplay.ColorHSV((1<<16)*2/3, 255, (actual%256)*24/256);
      } else
        color = BLACK;

      pixelsDisplay.setPixelColor(i, color);
    }
    pixelsDisplay.show();
  }

  if (changed || initial || display_timer == 0) {
    display_timer = 0;

    display.fillRect(0, 0, 128, 64, SSD1306_BLACK);
    display.drawBitmap(0, 0, logo_data, logo_width, logo_height, SSD1306_WHITE);

    display.setTextSize(1);
    display.setTextColor(SSD1306_WHITE);
    display.setCursor(0, 1);
    display.print(F("Soll: "));
    display.print(wantedTemperature/10);
    display.print(F("."));
    display.print(wantedTemperature%10);
    display.println();
    display.print(F("Ist:  "));
    display.print(actualTemperature/10);
    display.print(F("."));
    display.print(actualTemperature%10);
    display.println();

    if (wantedTemperatureUntil != TIME_DISABLED) {
      display.print(F("noch: "));
      int x = time_to_standby/1000/60/60;
      display.print(x);
      display.print(F(":"));
      x = time_to_standby/1000/60 % 60;
      if (x < 10)
        display.print("0");
      display.println(x);
    } else {
      display.println(F("noch: ----"));
    }

    display.print(F("int:  "));
    display.println(temp0, 1);

    //FIXME show as bar
    display.print(F("Heat: "));
    display.println(heatOn);

    //FIXME measure and display resistance of heating element (and whether there is one)

    if (DEBUG_TOUCH) {
      display.print(F("T:"));
      if (touch.available[0])
        displayPrintHexFixed(touch.touched[0], 4);
      else
        display.write("----");
      if (touch.available[1])
        displayPrintHexFixed(touch.touched[1], 4);
      else
        display.write("----");
      display.println();
    }

    if (DEBUG_TIME) {
      if (wantedTemperatureUntil == TIME_DISABLED)
        display.print(F("no timeout"));
      else
        display.print(time_to_standby);
      //display.println(buttons[3] ? ", X" : ", _");
      display.println();
    }

    display.setCursor(3, 57);
    display.println(F("---   +++   2h    +t"));

    display.display();
  }

  if (!digitalRead(BOOT2)) {
    analogWrite(LED_R, 128);
  } else {
    analogWrite(LED_R, 0);
  }

  delay(1);
}